<?php
/**
 * Author: Evgeny Zinoviev
 * Email: luethus@gmail.com
 * License: Apache License
 * June 4, 2011
 */

class Mrg {
	protected $_filename;
	public $levels = array(); // read-only
	protected $_levels_names = array('easy', 'medium', 'hard');
	
	public function __set($name, $value) {
		switch($name) {
		case 'levels':
			// read-only
			break;
		
		default:
			$this->$name = $value;
			break;
		}
	}
	
	public function open($filename, $load_all = false, $stop_on_error = false) {
		$f = fopen($filename, 'r');
		if(!$f) {
			return new Mrg_Error('Cannot open file: '.$filename);
		}
		
		$this->_filename = $filename;
		$filesize = filesize($filename);
		
		if(!$filesize) {
			fclose($f);
			return new Mrg_Error('File is empty');
		}
		
		for($i = 0; $i < 3; $i++) {
			$current_level = $this->_levels_names[$i];
			$this->levels[$current_level] = array();
			
			// read count of tracks
			$t = fread($f, 4);
			$count_tracks = $this->_unpackInt($t);
			
			if($stop_on_errors && $count_tracks < 0) {
				fclose($f);
				return new Mrg_Error('Negative value of tracks count, file is invalid');
			}
			
			$this->levels[$current_level] = new Mrg_Level;
			
			// read tracks
			for($j = 0; $j < $count_tracks; $j++) {
				if(feof($f)) {
					fclose($f);
					return new Mrg_Error('Unexpected end of file, file is invalid');
				}
				
				$track = new Mrg_Track;
				
				$t = fread($f, 4);
				$byte_addr = $this->_unpackInt($t);
				
				if($stop_on_error && $byte_addr < 0 || $byte_addr > $filesize) {
					fclose($f);
					return new Mrg_Error('Byte address of '.($j).' track in '.$current_level.' is invalid');
				}
				
				$track_name = '';
				$b = '';
				$c = 0;
				
				do {
					$b = fread($f, 1);
					$c = ord($b);
					
					if($c != 0) $track_name .= $b;
				} while($c != 0);
				
				$track->name = $track_name;
				$track->address = $byte_addr;
				
				$this->levels[$current_level]->addTrack($track);
			}
		}
		
		if($load_all) {
			// load all tracks to memory
			foreach($this->_levels_names as $_level) {
				$level = $this->getLevel($_level);
				
				for($i = 0; $i < $level->count; $i++) {
					$track = $level->getTrack($i);
					
					// set file position to track address
					fseek($f, $track->address);
					
					$data = $this->_readTrackData($f);
					if(!$data) {
						if($stop_on_error) {
							fclose($f);
							return new Mrg_Error('Cannot load track data, level='.$_level.', track='.$track->name);
						} else $track->removeAllPoints();
					} else {
						$track->setStart($data['start']->x, $data['start']->y);
						$track->setFinish($data['finish']->x, $data['finish']->y);
						$track->addPoints($data['points']);
					}
				}
			}
		}
		
		fclose($f);
		return true;
	}
	
	public function save($filename) {
		foreach($this->_levels_names as $_level) {
			if(empty($this->levels[$_level]) || !$this->levels[$_level]->count) {
				return new Mrg_Error('Empty '.$_level);
			}
		}
		
		$pos = 0;
		foreach($this->_levels_names as $_level) {
			$level = &$this->levels[$_level];
			$pos += 4;
			
			for($i = 0; $i < $level->count; $i++) {
				$track = &$level->tracks[$i];
				$pos += 4 + strlen($track->name) + 1;
			}
		}
		
		foreach($this->_levels_names as $_level) {
			$level = &$this->levels[$_level];
			
			for($i = 0; $i < $level->count; $i++) {
				$track = &$level->tracks[$i];
				$track->address = $pos;
				
				$pos += 27;
				$pos += ($track->count_points-1) * 2;
			}
		}
		
		$f = fopen($filename, 'w');
		if(!$f) {
			return new Mrg_Error('Cannot open file '.$filename.' for writing');
		}
		
		foreach($this->_levels_names as $_level) {
			$level = &$this->levels[$_level];
			fwrite($f, $this->_packInt($level->count));
			
			for($i = 0; $i < $level->count; $i++) {
				$track = &$level->tracks[$i];
				
				fwrite($f, $this->_packInt($track->address));
				fwrite($f, $track->name.chr(0));
			}
		}
		
		foreach($this->_levels_names as $_level) {
			$level = &$this->levels[$_level];
			for($i = 0; $i < $level->count; $i++) {
				$track = &$level->tracks[$i];
				
				fwrite($f, chr(0x33));
				fwrite($f, $this->_packInt($this->_packXYPosition($track->start->x)) . $this->_packInt($this->_packXYPosition($track->start->y)));
				fwrite($f, $this->_packInt($this->_packXYPosition($track->finish->x)) . $this->_packInt($this->_packXYPosition($track->finish->y)));
				fwrite($f, $this->_packShort($track->count_points));
				
				$first = $track->getPoint(0);
				fwrite($f, $this->_packInt($first->x) . $this->_packInt($first->y));
				
				$prev_point = $first;
				for($j = 1; $j < $track->count_points; $j++) {
					$point = $track->getPoint($j);
					
					$diff_x = $point->x - $prev_point->x;
					$diff_y = $point->y - $prev_point->y;
					
					if($j > 1) {
						if($diff_x > 127) $diff_x = 127;
						if($diff_y > 127) $diff_y = 127;
					}
					
					fwrite($f, chr($diff_x) . chr($diff_y));
					
					$prev_point = $point;
				}
			}
		}
		
		fclose($f);
		return true;
	}
	
	public function clean() {
		$this->levels = array();
		$this->_filename = '';
	}
	
	// returns a _reference_
	public function &getLevel($level) {
		$level = $this->_getLevel($level);
		if(!$level) {
			return new Mrg_Error('Unknown level '.$level);
		}
		
		if(empty($this->levels) || !isset($this->levels[$level])) {
			return new Mrg_Error('Level '.$level.' not found or mrg data is not loaded');
		}
		
		return $this->levels[$level];
	}
	
	public function loadTrack($level, $track) {
		if(!$this->_filename) {
			return Mrg_Error('No file loaded');
		}
		
		$level = $this->getLevel($level);
		if($level instanceof Mrg_Error) return $level;
		
		$track = $level->getTrack($track);
		if($track instanceof Mrg_Error) return $track;
		
		$f = fopen($this->_filename, 'r');
		if(!$f) {
			return Mrg_Error('Cannot open file: '.$this->_filename);
		}
		
		fseek($f, $track->address);
		
		if(feof($f)) {
			fclose($f);
			return Mrg_Error('Invalid track address in file, cannot load data');
		}
		
		$data = $this->_readTrackData($f);
		fclose($f);
		if($data === false) {
			return Mrg_Error('Something is not right, maybe corrupted file...');
		}
		
		$track->setStart($data['start']);
		$track->setFinish($data['finish']);
		
		$track->removeAllPoints();
		$track->addPoints($data['points']);
		
		return $track;
	}
	
	/** protected methods **/
	protected function _getLevel($level) {
		if(is_int($level)) {
			if(isset($this->_levels_names[$level])) $level = $this->_levels_names[$level];
			else {
				return false;
			}
		} else {
			$level = strtolower($level);
			if(!in_array($level, $this->_levels_names)) {
				return false;
			}
		}
		
		return $level;
	}
	
	protected function _readTrackData(&$f) {
		$b33 = ord(fread($f, 1));
		if($b33 != 0x33) return false;
		
		// start position of bike
		$start = new Mrg_Point;
		foreach(array('x', 'y') as $pt) $start->$pt = $this->_unpackXYPosition($this->_unpackInt(fread($f, 4)));
		
		// finish position (flag)
		$finish = new Mrg_Point;
		foreach(array('x', 'y') as $pt) $finish->$pt = $this->_unpackXYPosition($this->_unpackInt(fread($f, 4)));
		
		// short
		$count_points = $this->_unpackShort(fread($f, 2));
		
		$point = new Mrg_Point;
		
		$cX = $this->_unpackInt(fread($f, 4));
		$cY = $this->_unpackInt(fread($f, 4));
		
		$point->x = $cX;
		$point->y = $cY;
		
		$points[] = $point;
		
		// points
		for($j = 1; $j < $count_points; $j++) {
			$point = new Mrg_Point;
			
			$b1 = $this->_getSignedPoint(ord(fread($f, 1)));
			if($b1 == -1) {
				$cX = 0;
				$x = $this->_unpackInt(fread($f, 4));
				$y = $this->_unpackInt(fread($f, 4));
				
				//if(abs($x) > 127) $x = 0;
				if(abs($y) > 127) $y = 0;
			} else {
				if($b1 == 255) $b1 = 0;
				
				$x = $b1;
				$y = $this->_getSignedPoint(ord(fread($f, 1)));
			}
			
			$cX += $x;
			$cY += $y;
			
			$point->x = $cX;
			$point->y = $cY;
			
			$points[] = $point;
		}
		
		return array (
			'start' => $start,
			'finish' => $finish,
			'points' => $points
		);
	}
	
	protected function _unpackInt($str) {
		return (ord($str{0}) << 24) | (ord($str{1}) << 16) | (ord($str{2}) << 8) | ord($str{3});
	}
	
	protected function _packInt($int) {
		$str = chr(($int >> 24) & 0xff);
		$str .= chr(($int >> 16) & 0xff);
		$str .= chr(($int >> 8) & 0xff);
		$str .= chr($int & 0xff);
		
		return $str;
	}
	
	protected function _unpackShort($str) {
		return (ord($str{0}) << 8) | ord($str{1});
	}
	
	protected function _packShort($short) {
		$str = chr(($short >> 8) & 0xff);
		$str .= chr($short & 0xff);
		
		return $str;
	}
	
	protected function _unpackXYPosition($int) {
		return ($int >> 16) << 3;
	}
	
	protected function _packXYPosition($int) {
		return ($int << 16) >> 3;
	}
	
	protected function _getSignedPoint($int) {
		if($int > 127) $int = -(256 - $int);
		return $int;
	}
}

class Mrg_Level {
	public $tracks; // array of Mrg_Tracks objects
	public $count; // int 
	
	public function __set($name, $value) {
		switch($name) {
		case 'count':
		case 'tracks':
			break;
			
		default:
			$this->$name = $value;
			break;
		}
	}
	
	public function addTrack(Mrg_Track $track) {
		$this->tracks[] = $track;
		$this->_recountTracks();
		
		return $this->count-1;
	}
	
	public function &getTrack($track, $clone = false) {
		$tmp = $track;
		$track = $this->_getTrack($track);
		
		if($track !== false) return (!$clone ? $this->tracks[$track] : clone $this->tracks[$track]);
		else return new Mrg_Error('Track '.$tmp.' not found');
	}

	public function removeTrack($index) {
		$index = $this->_getTrack($index);
		if($index !== false) {
			array_splice($this->tracks, $index, 1);
			$this->_recountTracks();
		}
	}
	
	public function removeAllTracks() {
		$this->tracks = array();
		$this->_recountTracks();
	}
	
	public function reorder($track, $pos, $swap = false) {
		$index = $this->_getTrack($track);
		if($index === false) return new Mrg_Error('Track '.$index.' does not exists');
		
		$max_pos = $this->count-1;
		
		if($pos < 0) $pos = 0;
		if($pos > $max_pos) $pos = $max_pos;
		
		if($swap) {
			$tmp = $this->tracks[$index];
			$this->tracks[$index] = $this->tracks[$pos];
			$this->tracks[$pos] = $tmp;
		} else {
			$track = $this->tracks[$index];
			array_splice($this->tracks, $index, 1);
			
			$size = sizeof($this->tracks);
			for($i = $size; $i >= 0; $i--) {
				if($i == $pos) {
					$this->tracks[$i] = $track;
					break;
				}
				
				if($i > 0) $this->tracks[$i] = $this->tracks[$i-1];
			}
		}
		
		$this->_recountTracks();
	}
	
	public function getList() {
		$list = array();
		for($i = 0; $i < $this->count; $i++) {
			$list[] = $this->tracks[$i]->name;
		}
		
		return $list;
	}
	
	protected function _recountTracks() {
		$this->count = sizeof($this->tracks);
	}
	
	protected function _getTrack($track) {
		$track_index = false;
		if(is_int($track)) {
			if(isset($this->tracks[$track]))
				$track_index = $track;
		} else {
			foreach($this->tracks as $_index => $_track) {
				if($_track->name == $track) {
					$track_index = $_index;
					break;
				}
			}
		}
		
		return $track_index;
	}
}

class Mrg_Track {
	public $name; // string
	public $address; // int
	
	// read-only
	public $count_points; // int
	public $points; // array of Mrg_Point objects
	public $start; // Mrg_Point
	public $finish; // Mrg_Point
	
	public function __get($name) {
		return isset($this->$name) ? $this->$name : null;
	}
	
	public function __set($name, $value) {
		switch($name) {
		case 'name':
			$this->$name = (string)$value;
			break;
		
		case 'count_points':
		case 'points':
		case 'start':
		case 'finish':
			// hahaha nonono...
			break;
		
		default:
			$this->$name = $value;
			break;
		}
	}
	
	public function getPoints() {
		return $this->points;
	}
	
	public function getPoint($index) {
		if(!isset($this->points[$index])) 
			return new Mrg_Error('Point '.$index.' not found');
		
		return $this->points[$index];
	}
	
	public function addPoint($x, $y) {
		$point = new Mrg_Point($x, $y);
		
		$end = false;
		if($this->count_points) $end = end($this->points);
		
		if(!$end || $end->x <= $x) $this->points[] = $point;
		else {
			$size = sizeof($this->points);
			$prev = false;
			
			for($i = 0; $i < $size; $i++) {
				if($this->points[$i]->x <= $x) continue;
				else {
					$prev = $i-1;
					break;
				}
			}
			
			if($prev !== false) {
				for($i = $size; $i >= $prev; $i--) {
					if($i == $prev+1) {
						$this->points[$i] = $point;
						break;
					}
					
					$this->points[$i] = $this->points[$i-1];
				}
			} else $this->points[] = $point;
		}
		
		$this->_recountPoints();
	}
	
	public function addPoints(array $points) {
		foreach($points as $point) $this->addPoint($point->x, $point->y);
	}
	
	public function setStart($x, $y) {
		$this->start = new Mrg_Point($x, $y);
	}
	
	public function setFinish($x) {
		$this->finish = new Mrg_Point($x, 0);
	}
	
	public function editPoint($index, $x, $y) {
		if(!isset($this->points[$index])) return new Mrg_Error('Point '.$index.' not found');
		
		$this->points[$index]->x = $x;
		$this->points[$index]->y = $y;
	}
	
	public function removePoint($index) {
		if(!isset($this->points[$index])) return Mrg_Error('Point '.$index.' does not exists');
		
		array_splice($this->points, $index, 1);
		$this->_recountPoints();
	}
	
	public function removeAllPoints() {
		$this->count_points = 0;
		$this->points = array();
	}
	
	public function getImage($width = null) {
		if(!$this->count_points)
			return new Mrg_Error('Track data not loaded');
		
		if(!$width || $width < 0) $width = null;
		
		$x = array(0);
		$y = array(0);
		
		$first = $this->getPoint(0);
		
		for($i = 1; $i < $this->count_points; $i++) {
			$point = $this->getPoint($i);
			
			$x[] = $point->x - $first->x;
			$y[] = - ($point->y - $first->y);
		}
		
		if($width == null) $width = max($x);
		$scale = max($x) / $width;
		
		$_x = (max($y) - min($y)) - max($y);
		
		$w = max($x);
		$h = max($y)-min($y);
		
		$img = imagecreatetruecolor(max($x)/$scale, (max($y)-min($y))/$scale+1);
		
		$white = imagecolorallocate($img, 255, 255, 255);
		$green = imagecolorallocate($img, 0, 170, 0);
		
		imagefilledrectangle ($img, 0, 0, $w, $h, $white);
		
		$count = sizeof($x);
		for($i = 1; $i < $count; $i++)
			imageline($img, abs($x[$i-1])/$scale, ($y[$i-1]+$_x)/$scale, abs($x[$i])/$scale, ($y[$i]+$_x)/$scale, $green);
			
		return $img;
	}
	
	protected function _recountPoints() {
		$this->count_points = sizeof($this->points);
	}
}

class Mrg_Point {
	public $x; // int
	public $y; // int
	
	public function __set($name, $value) {
		switch($name) {
		case 'x':
		case 'y':
			$this->$name = (int)$value;
			break;
			
		default:
			$this->$name = $value;
			break;
		}
	}
	
	public function __construct($x = null, $y = null) {
		if($x === null) return;
		
		$this->x = (int)$x;
		$this->y = (int)$y;
	}
}

class Mrg_Error {
	public $error;
	
	public function __construct($error) {
		$this->error = $error;
	}
}
?>
